import unittest

from test.boggle_test import BoggleTest
from test.scrabble_test import ScrabbleTest
from test.hangman_test import HangmanTest
from test.wordsearch_test import WordSearchTest

if __name__ == "__main__":
    unittest.main()
