"""Wordsolver Module"""

__version__ = "1.0.0"

from wordsolver.boggle import BoggleSolver
from wordsolver.scrabble import ScrabbleSolver, VALUE_STANDARD, PREMIUM_STANDARD, EMPTY_STANDARD
from wordsolver.hangman import HangmanSolver
from wordsolver.wordsearch import WordSearchSolver
